@@ -1,13 +1,14 @@
 import asyncio
 import enum
-import hashlib
 import ipaddress
 import logging
+import random
 import socket
 
 import netifaces
 
 from . import exceptions, stun, turn
+from .candidate import Candidate, candidate_foundation, candidate_priority
 from .compat import secrets
 from .utils import random_string
 
@@ -16,31 +17,8 @@ logger = logging.getLogger('ice')
 ICE_COMPLETED = 1
 ICE_FAILED = 2
 
-
-def candidate_foundation(candidate_type, candidate_transport, base_address):
-    """
-    See RFC 5245 - 4.1.1.3. Computing Foundations
-    """
-    key = '%s|%s|%s' % (type, candidate_transport, base_address)
-    return hashlib.md5(key.encode('ascii')).hexdigest()
-
-
-def candidate_priority(candidate_component, candidate_type, local_pref=65535):
-    """
-    See RFC 5245 - 4.1.2.1. Recommended Formula
-    """
-    if candidate_type == 'host':
-        type_pref = 126
-    elif candidate_type == 'prflx':
-        type_pref = 110
-    elif candidate_type == 'srflx':
-        type_pref = 100
-    else:
-        type_pref = 0
-
-    return (1 << 24) * type_pref + \
-           (1 << 8) * local_pref + \
-           (256 - candidate_component)
+CONSENT_FAILURES = 6
+CONSENT_INTERVAL = 5
 
 
 def candidate_pair_priority(local, remote, ice_controlling):
@@ -99,46 +77,6 @@ def sort_candidate_pairs(pairs, ice_controlling):
     pairs.sort(key=pair_priority)
 
 
-class Candidate:
-    """
-    An ICE candidate.
-    """
-    def __init__(self, foundation, component, transport, priority, host, port,
-                 type='host', generation=0):
-        self.foundation = foundation
-        self.component = component
-        self.transport = transport
-        self.priority = priority
-        self.host = host
-        self.port = port
-        self.type = type
-        self.generation = generation
-
-    def __repr__(self):
-        return 'Candidate(%s)' % self
-
-    def __str__(self):
-        return '%s %d %s %d %s %d typ %s generation %d' % (
-            self.foundation,
-            self.component,
-            self.transport,
-            self.priority,
-            self.host,
-            self.port,
-            self.type,
-            self.generation)
-
-    def can_pair_with(self, other):
-        """
-        A local candidate is paired with a remote candidate if and only if
-        the two candidates have the same component ID and have the same IP
-        address version.
-        """
-        a = ipaddress.ip_address(self.host)
-        b = ipaddress.ip_address(other.host)
-        return self.component == other.component and a.version == b.version
-
-
 class CandidatePair:
     def __init__(self, protocol, remote_candidate):
         self.handle = None
@@ -175,19 +113,6 @@ class CandidatePair:
         FAILED = 4
 
 
-def parse_candidate(value):
-    bits = value.split()
-    return Candidate(
-        foundation=bits[0],
-        component=int(bits[1]),
-        transport=bits[2],
-        priority=int(bits[3]),
-        host=bits[4],
-        port=int(bits[5]),
-        type=bits[7],
-        generation=int(bits[9]))
-
-
 def next_protocol_id():
     protocol_id = next_protocol_id.counter
     next_protocol_id.counter += 1
@@ -208,6 +133,7 @@ class StunProtocol(asyncio.DatagramProtocol):
 
     def connection_lost(self, exc):
         self.__log_debug('connection_lost(%s)', exc)
+        asyncio.ensure_future(self.queue.put(None))
         self.__closed.set_result(True)
 
     def connection_made(self, transport):
@@ -220,7 +146,7 @@ class StunProtocol(asyncio.DatagramProtocol):
 
         try:
             message = stun.parse_message(data)
-            self.__log_debug('< %s %s', addr, repr(message))
+            self.__log_debug('< %s %s', addr, message)
         except ValueError:
             self.__log_debug('< %s DATA %d', addr, len(data))
             coro = self.queue.put(data)
@@ -248,10 +174,10 @@ class StunProtocol(asyncio.DatagramProtocol):
         return await self.queue.get(), self.local_candidate.component
 
     async def send_data(self, data, addr):
-        self.__log_debug('%s DATA %d', addr, len(data))
+        self.__log_debug('> %s DATA %d', addr, len(data))
         self.transport.sendto(data, addr)
 
-    async def request(self, request, addr, integrity_key=None):
+    async def request(self, request, addr, integrity_key=None, retransmissions=None):
         """
         Execute a STUN transaction and return the response.
         """
@@ -261,7 +187,7 @@ class StunProtocol(asyncio.DatagramProtocol):
             request.add_message_integrity(integrity_key)
             request.add_fingerprint()
 
-        transaction = stun.Transaction(request, addr, self)
+        transaction = stun.Transaction(request, addr, self, retransmissions=retransmissions)
         transaction.integrity_key = integrity_key
         self.transactions[request.transaction_id] = transaction
         try:
@@ -273,11 +199,11 @@ class StunProtocol(asyncio.DatagramProtocol):
         """
         Send a STUN message.
         """
-        self.__log_debug('> %s %s', addr, repr(message))
+        self.__log_debug('> %s %s', addr, message)
         self.transport.sendto(bytes(message), addr)
 
     def __log_debug(self, msg, *args):
-        logger.debug(repr(self.receiver) + ' ' + repr(self) + ' ' + msg, *args)
+        logger.debug('%s %s ' + msg, self.receiver, self, *args)
 
     def __repr__(self):
         return 'protocol(%s)' % self.id
@@ -294,47 +220,118 @@ next_connection_id.counter = 0
 
 class Connection:
     """
-    An ICE connection.
+    An ICE connection for a single media stream.
     """
-    def __init__(self, ice_controlling, stun_server=None,
-                 turn_server=None, turn_username=None, turn_password=None,
-                 use_ipv4=True, use_ipv6=False):
-        self.components = set([1])
+    def __init__(self, ice_controlling, components=1,
+                 stun_server=None,
+                 turn_server=None, turn_username=None, turn_password=None, turn_transport='udp',
+                 use_ipv4=True, use_ipv6=True):
         self.ice_controlling = ice_controlling
-        self.id = next_connection_id()
+        #: Local username, automatically set to a random value.
         self.local_username = random_string(4)
+        #: Local password, automatically set to a random value.
         self.local_password = random_string(22)
-        self.remote_candidates = []
+        #: Remote username, which you need to set.
         self.remote_username = None
+        #: Remote password, which you need to set.
         self.remote_password = None
         self.stun_server = stun_server
-        self.tie_breaker = secrets.randbits(64)
         self.turn_server = turn_server
         self.turn_username = turn_username
         self.turn_password = turn_password
+        self.turn_transport = turn_transport
 
         # private
-        self.__nominated = {}
-        self.addresses = get_host_addresses(use_ipv4=use_ipv4, use_ipv6=use_ipv6)
-        self.check_list = []
-        self.check_list_state = asyncio.Queue()
-        self.early_checks = []
-        self.protocols = []
+        self._components = set(range(1, components + 1))
+        self._check_list = []
+        self._check_list_done = False
+        self._check_list_state = asyncio.Queue()
+        self._early_checks = []
+        self._id = next_connection_id()
+        self._local_candidates = []
+        self._local_candidates_end = False
+        self._local_candidates_start = False
+        self._nominated = {}
+        self._protocols = []
+        self._received = []
+        self._remote_candidates = []
+        self._remote_candidates_end = False
+        self._query_consent_handle = None
+        self._tie_breaker = secrets.randbits(64)
+        self._use_ipv4 = use_ipv4
+        self._use_ipv6 = use_ipv6
 
-    async def get_local_candidates(self):
+    @property
+    def local_candidates(self):
+        """
+        Local candidates, automatically set by :meth:`gather_candidates`.
+        """
+        return self._local_candidates[:]
+
+    @property
+    def remote_candidates(self):
+        """
+        Remote candidates, which you need to set.
+
+        Assigning this attribute will automatically signal end-of-candidates.
+        If you will be adding more remote candidates in the future, use the
+        :meth:`add_remote_candidate` method instead.
+        """
+        return self._remote_candidates[:]
+
+    @remote_candidates.setter
+    def remote_candidates(self, value):
+        if self._remote_candidates_end:
+            raise ValueError('Cannot set remote candidates after end-of-candidates.')
+
+        self._remote_candidates = value[:]
+        self._prune_components()
+        self._remote_candidates_end = True
+
+    def add_remote_candidate(self, remote_candidate):
+        """
+        Add a remote candidate or signal end-of-candidates.
+
+        To signal end-of-candidates, pass `None`.
+        """
+        if self._remote_candidates_end:
+            raise ValueError('Cannot add remote candidate after end-of-candidates.')
+
+        if remote_candidate is None:
+            self._prune_components()
+            self._remote_candidates_end = True
+            return
+
+        self._remote_candidates.append(remote_candidate)
+        for protocol in self._protocols:
+            if (protocol.local_candidate.can_pair_with(remote_candidate) and
+               not self._find_pair(protocol, remote_candidate)):
+                pair = CandidatePair(protocol, remote_candidate)
+                self._check_list.append(pair)
+        self.sort_check_list()
+
+    async def gather_candidates(self):
         """
         Gather local candidates.
-        """
-        candidates = []
-        for component in self.components:
-            candidates += await self.get_component_candidates(component)
-        return candidates
 
-    def set_remote_candidates(self, candidates):
+        You **must** call this coroutine before calling :meth:`connect`.
         """
-        Set remote candidates.
+        if not self._local_candidates_start:
+            self._local_candidates_start = True
+            addresses = get_host_addresses(use_ipv4=self._use_ipv4, use_ipv6=self._use_ipv6)
+            for component in self._components:
+                self._local_candidates += await self.get_component_candidates(
+                    component=component,
+                    addresses=addresses)
+            self._local_candidates_end = True
+
+    def get_default_candidate(self, component):
         """
-        self.remote_candidates = candidates
+        Gets the default local candidate for the specified component.
+        """
+        for candidate in sorted(self._local_candidates, key=lambda x: x.priority):
+            if candidate.component == component:
+                return candidate
 
     async def connect(self):
         """
@@ -343,43 +340,28 @@ class Connection:
         This coroutine returns if a candidate pair was successfuly nominated
         and raises an exception otherwise.
         """
+        if not self._local_candidates_end:
+            raise ConnectionError('Local candidates gathering was not performed')
+
         if (self.remote_username is None or
            self.remote_password is None):
-            raise exceptions.ImproperlyConfigured('Remote username or password is missing')
+            raise ConnectionError('Remote username or password is missing')
 
         # 5.7.1. Forming Candidate Pairs
-        for remote_candidate in self.remote_candidates:
-            for protocol in self.protocols:
-                if protocol.local_candidate.can_pair_with(remote_candidate):
+        for remote_candidate in self._remote_candidates:
+            for protocol in self._protocols:
+                if (protocol.local_candidate.can_pair_with(remote_candidate) and
+                   not self._find_pair(protocol, remote_candidate)):
                     pair = CandidatePair(protocol, remote_candidate)
-                    self.check_list.append(pair)
+                    self._check_list.append(pair)
         self.sort_check_list()
-        if not self.check_list:
-            raise exceptions.ConnectionError('No candidate pairs formed')
 
-        # unfreeze first pair for component 1
-        first_pair = None
-        for pair in self.check_list:
-            if pair.component == 1:
-                first_pair = pair
-                break
-        assert first_pair is not None
-        if first_pair.state == CandidatePair.State.FROZEN:
-            self.check_state(first_pair, CandidatePair.State.WAITING)
-
-        # unfreeze pairs with same component but different foundations
-        seen_foundations = set(first_pair.local_candidate.foundation)
-        for pair in self.check_list:
-            if (pair.component == first_pair.component and
-               pair.local_candidate.foundation not in seen_foundations and
-               pair.state == CandidatePair.State.FROZEN):
-                self.check_state(pair, CandidatePair.State.WAITING)
-                seen_foundations.add(pair.local_candidate.foundation)
+        self._unfreeze_initial()
 
         # handle early checks
-        for check in self.early_checks:
+        for check in self._early_checks:
             self.check_incoming(*check)
-        self.early_checks = []
+        self._early_checks = []
 
         # perform checks
         while True:
@@ -388,29 +370,51 @@ class Connection:
             await asyncio.sleep(0.02)
 
         # wait for completion
-        res = await self.check_list_state.get()
+        if self._check_list:
+            res = await self._check_list_state.get()
+        else:
+            res = ICE_FAILED
 
         # cancel remaining checks
-        for check in self.check_list:
+        for check in self._check_list:
             if check.handle:
                 check.handle.cancel()
 
         if res != ICE_COMPLETED:
-            raise exceptions.ConnectionError
+            raise ConnectionError('ICE negotiation failed')
+
+        # start consent freshness tests
+        self._query_consent_handle = asyncio.ensure_future(self.query_consent())
 
     async def close(self):
         """
         Close the connection.
         """
-        for protocol in self.protocols:
+        # stop consent freshness tests
+        if self._query_consent_handle and not self._query_consent_handle.done():
+            self._query_consent_handle.cancel()
+            try:
+                await self._query_consent_handle
+            except asyncio.CancelledError:
+                pass
+
+        # stop check list
+        if self._check_list and not self._check_list_done:
+            await self._check_list_state.put(ICE_FAILED)
+
+        self._nominated.clear()
+        for protocol in self._protocols:
             await protocol.close()
-        self.protocols = []
+        self._protocols.clear()
+        self._local_candidates.clear()
 
     async def recv(self):
         """
         Receive the next datagram.
 
         The return value is a `bytes` object representing the data received.
+
+        If the connection is not established, a `ConnectionError` is raised.
         """
         data, component = await self.recvfrom()
         return data
@@ -422,43 +426,97 @@ class Connection:
         The return value is a `(bytes, component)` tuple where `bytes` is a
         bytes object representing the data received and `component` is the
         component on which the data was received.
+
+        If the connection is not established, a `ConnectionError` is raised.
         """
-        fs = [protocol.recv_data() for protocol in self.protocols]
-        done, pending = await asyncio.wait(fs, return_when=asyncio.FIRST_COMPLETED)
-        for task in pending:
-            task.cancel()
-        assert len(done) == 1
-        return done.pop().result()
+        if not len(self._nominated):
+            raise ConnectionError('Cannot receive data, not connected')
+
+        if not self._received:
+            fs = [protocol.recv_data() for protocol in self._protocols]
+            done, pending = await asyncio.wait(fs, return_when=asyncio.FIRST_COMPLETED)
+            for task in pending:
+                task.cancel()
+            for task in done:
+                result = task.result()
+                if result[0] is None:
+                    raise ConnectionError('Connection lost while receiving data')
+                self._received.append(result)
+
+        return self._received.pop(0)
 
     async def send(self, data):
         """
         Send a datagram on the first component.
+
+        If the connection is not established, a `ConnectionError` is raised.
         """
         await self.sendto(data, 1)
 
     async def sendto(self, data, component):
         """
         Send a datagram on the specified component.
+
+        If the connection is not established, a `ConnectionError` is raised.
         """
-        active_pair = self.__nominated.get(component)
+        active_pair = self._nominated.get(component)
         if active_pair:
             await active_pair.protocol.send_data(data, active_pair.remote_addr)
+        else:
+            raise ConnectionError('Cannot send data, not connected')
+
+    def set_selected_pair(self, component, local_foundation, remote_foundation):
+        """
+        Force the selected candidate pair.
+
+        If the remote party does not support ICE, you should using this
+        instead of calling :meth:`connect`.
+        """
+        # find local candidate
+        protocol = None
+        for p in self._protocols:
+            if (p.local_candidate.component == component and
+               p.local_candidate.foundation == local_foundation):
+                protocol = p
+                break
+
+        # find remote candidate
+        remote_candidate = None
+        for c in self._remote_candidates:
+            if c.component == component and c.foundation == remote_foundation:
+                remote_candidate = c
+
+        assert (protocol and remote_candidate)
+        self._nominated[component] = CandidatePair(protocol, remote_candidate)
 
     # private
 
+    def build_request(self, pair):
+        tx_username = '%s:%s' % (self.remote_username, self.local_username)
+        request = stun.Message(message_method=stun.Method.BINDING,
+                               message_class=stun.Class.REQUEST)
+        request.attributes['USERNAME'] = tx_username
+        request.attributes['PRIORITY'] = candidate_priority(pair.component, 'prflx')
+        if self.ice_controlling:
+            request.attributes['ICE-CONTROLLING'] = self._tie_breaker
+            request.attributes['USE-CANDIDATE'] = None
+        else:
+            request.attributes['ICE-CONTROLLED'] = self._tie_breaker
+        return request
+
     def check_complete(self, pair):
         pair.handle = None
 
         if pair.state == CandidatePair.State.SUCCEEDED:
             if pair.nominated:
-                self.__nominated[pair.component] = pair
+                self._nominated[pair.component] = pair
 
                 # 8.1.2.  Updating States
                 #
                 # The agent MUST remove all Waiting and Frozen pairs in the check
                 # list and triggered check queue for the same component as the
                 # nominated pairs for that media stream.
-                for p in self.check_list:
+                for p in self._check_list:
                     if (p.component == pair.component and
                        p.state in [CandidatePair.State.WAITING, CandidatePair.State.FROZEN]):
                         self.check_state(p, CandidatePair.State.FAILED)
@@ -466,28 +524,32 @@ class Connection:
             # Once there is at least one nominated pair in the valid list for
             # every component of at least one media stream and the state of the
             # check list is Running:
-            if len(self.__nominated) == len(self.components):
-                self.__log_info('ICE completed')
-                asyncio.ensure_future(self.check_list_state.put(ICE_COMPLETED))
+            if len(self._nominated) == len(self._components):
+                if not self._check_list_done:
+                    self.__log_info('ICE completed')
+                    asyncio.ensure_future(self._check_list_state.put(ICE_COMPLETED))
+                    self._check_list_done = True
                 return
 
             # 7.1.3.2.3.  Updating Pair States
-            for p in self.check_list:
+            for p in self._check_list:
                 if (p.local_candidate.foundation == pair.local_candidate.foundation and
                    p.state == CandidatePair.State.FROZEN):
                     self.check_state(p, CandidatePair.State.WAITING)
 
-        for p in self.check_list:
+        for p in self._check_list:
             if p.state not in [CandidatePair.State.SUCCEEDED, CandidatePair.State.FAILED]:
                 return
 
         if not self.ice_controlling:
-            for p in self.check_list:
+            for p in self._check_list:
                 if p.state == CandidatePair.State.SUCCEEDED:
                     return
 
-        self.__log_info('ICE failed')
-        asyncio.ensure_future(self.check_list_state.put(ICE_FAILED))
+        if not self._check_list_done:
+            self.__log_info('ICE failed')
+            asyncio.ensure_future(self._check_list_state.put(ICE_FAILED))
+            self._check_list_done = True
 
     def check_incoming(self, message, addr, protocol):
         """
@@ -497,7 +559,7 @@ class Connection:
 
         # find remote candidate
         remote_candidate = None
-        for c in self.remote_candidates:
+        for c in self._remote_candidates:
             if c.host == addr[0] and c.port == addr[1]:
                 remote_candidate = c
                 assert remote_candidate.component == component
@@ -512,19 +574,15 @@ class Connection:
                 host=addr[0],
                 port=addr[1],
                 type='prflx')
-            self.remote_candidates.append(remote_candidate)
-            self.__log_info('Discovered peer reflexive candidate %s' % repr(remote_candidate))
+            self._remote_candidates.append(remote_candidate)
+            self.__log_info('Discovered peer reflexive candidate %s', remote_candidate)
 
         # find pair
-        pair = None
-        for p in self.check_list:
-            if (p.protocol == protocol and p.remote_candidate == remote_candidate):
-                pair = p
-                break
+        pair = self._find_pair(protocol, remote_candidate)
         if pair is None:
             pair = CandidatePair(protocol, remote_candidate)
             pair.state = CandidatePair.State.WAITING
-            self.check_list.append(pair)
+            self._check_list.append(pair)
             self.sort_check_list()
 
         # triggered check
@@ -541,17 +599,21 @@ class Connection:
 
     def check_periodic(self):
         # find the highest-priority pair that is in the waiting state
-        for pair in self.check_list:
+        for pair in self._check_list:
             if pair.state == CandidatePair.State.WAITING:
                 pair.handle = asyncio.ensure_future(self.check_start(pair))
                 return True
 
         # find the highest-priority pair that is in the frozen state
-        for pair in self.check_list:
+        for pair in self._check_list:
             if pair.state == CandidatePair.State.FROZEN:
                 pair.handle = asyncio.ensure_future(self.check_start(pair))
                 return True
 
+        # if we expect more candidates, keep going
+        if not self._remote_candidates_end:
+            return not self._check_list_done
+
         return False
 
     async def check_start(self, pair):
@@ -560,17 +622,7 @@ class Connection:
         """
         self.check_state(pair, CandidatePair.State.IN_PROGRESS)
 
-        tx_username = '%s:%s' % (self.remote_username, self.local_username)
-        request = stun.Message(message_method=stun.Method.BINDING,
-                               message_class=stun.Class.REQUEST)
-        request.attributes['USERNAME'] = tx_username
-        request.attributes['PRIORITY'] = candidate_priority(pair.component, 'prflx')
-        if self.ice_controlling:
-            request.attributes['ICE-CONTROLLING'] = self.tie_breaker
-            request.attributes['USE-CANDIDATE'] = None
-        else:
-            request.attributes['ICE-CONTROLLED'] = self.tie_breaker
-
+        request = self.build_request(pair)
         try:
             response, addr = await pair.protocol.request(
                 request, pair.remote_addr,
@@ -590,7 +642,7 @@ class Connection:
 
         # check remote address matches
         if addr != pair.remote_addr:
-            self.__log_info('Check %s failed : source address mismatch' % repr(pair))
+            self.__log_info('Check %s failed : source address mismatch', pair)
             self.check_state(pair, CandidatePair.State.FAILED)
             self.check_complete(pair)
             return
@@ -605,19 +657,32 @@ class Connection:
         """
         Updates the state of a check.
         """
-        self.__log_info('Check %s %s -> %s' % (repr(pair), pair.state, state))
+        self.__log_info('Check %s %s -> %s', pair, pair.state, state)
         pair.state = state
 
-    async def get_component_candidates(self, component, timeout=5):
+    def _find_pair(self, protocol, remote_candidate):
+        """
+        Find a candidate pair in the check list.
+        """
+        for pair in self._check_list:
+            if (pair.protocol == protocol and pair.remote_candidate == remote_candidate):
+                return pair
+        return None
+
+    async def get_component_candidates(self, component, addresses, timeout=5):
         candidates = []
 
         loop = asyncio.get_event_loop()
-        for address in self.addresses:
+        for address in addresses:
             # create transport
-            _, protocol = await loop.create_datagram_endpoint(
-                lambda: StunProtocol(self),
-                local_addr=(address, 0))
-            self.protocols.append(protocol)
+            try:
+                _, protocol = await loop.create_datagram_endpoint(
+                    lambda: StunProtocol(self),
+                    local_addr=(address, 0))
+            except OSError as exc:
+                self.__log_info('Could not bind to %s - %s', address, exc)
+                continue
+            self._protocols.append(protocol)
 
             # add host candidate
             candidate_address = protocol.transport.get_extra_info('sockname')
@@ -635,7 +700,7 @@ class Connection:
         if self.stun_server:
             # we query STUN server for IPv4
             fs = []
-            for protocol in self.protocols:
+            for protocol in self._protocols:
                 if ipaddress.ip_address(protocol.local_candidate.host).version == 4:
                     fs.append(server_reflexive_candidate(protocol, self.stun_server))
             if len(fs):
@@ -651,15 +716,16 @@ class Connection:
                 lambda: StunProtocol(self),
                 server_addr=self.turn_server,
                 username=self.turn_username,
-                password=self.turn_password)
-            self.protocols.append(protocol)
+                password=self.turn_password,
+                transport=self.turn_transport)
+            self._protocols.append(protocol)
 
             # add relayed candidate
             candidate_address = protocol.transport.get_extra_info('sockname')
             protocol.local_candidate = Candidate(
-                foundation=candidate_foundation('relay', 'udp', candidate_address[0]),
+                foundation=candidate_foundation('relay', self.turn_transport, candidate_address[0]),
                 component=component,
-                transport='udp',
+                transport=self.turn_transport,
                 priority=candidate_priority(component, 'relay'),
                 host=candidate_address[0],
                 port=candidate_address[1],
@@ -668,6 +734,42 @@ class Connection:
 
         return candidates
 
+    def _prune_components(self):
+        """
+        Remove components for which the remote party did not provide any candidates.
+
+        This can only be determined after end-of-candidates.
+        """
+        seen_components = set(map(lambda x: x.component, self._remote_candidates))
+        missing_components = self._components - seen_components
+        if missing_components:
+            self.__log_info('Components %s have no candidate pairs' % missing_components)
+            self._components = seen_components
+
+    async def query_consent(self):
+        """
+        Periodically check consent (RFC 7675).
+        """
+        failures = 0
+        while True:
+            # randomize between 0.8 and 1.2 times CONSENT_INTERVAL
+            await asyncio.sleep(CONSENT_INTERVAL * (0.8 + 0.4 * random.random()))
+
+            for pair in self._nominated.values():
+                request = self.build_request(pair)
+                try:
+                    await pair.protocol.request(
+                        request, pair.remote_addr,
+                        integrity_key=self.remote_password.encode('utf8'),
+                        retransmissions=0)
+                    failures = 0
+                except exceptions.TransactionError:
+                    failures += 1
+                if failures >= CONSENT_FAILURES:
+                    self.__log_info('Consent to send expired')
+                    self._query_consent_handle = None
+                    return await self.close()
+
     def request_received(self, message, addr, protocol, raw_data):
         if message.message_method != stun.Method.BINDING:
             self.respond_error(message, addr, protocol, (400, 'Bad Request'))
@@ -677,9 +779,10 @@ class Connection:
         try:
             stun.parse_message(raw_data,
                                integrity_key=self.local_password.encode('utf8'))
-            rx_username = '%s:%s' % (self.local_username, self.remote_username)
-            if message.attributes.get('USERNAME') != rx_username:
-                raise ValueError('Wrong username')
+            if self.remote_username is not None:
+                rx_username = '%s:%s' % (self.local_username, self.remote_username)
+                if message.attributes.get('USERNAME') != rx_username:
+                    raise ValueError('Wrong username')
         except ValueError as exc:
             self.respond_error(message, addr, protocol, (400, 'Bad Request'))
             return
@@ -687,13 +790,13 @@ class Connection:
         # 7.2.1.1. Detecting and Repairing Role Conflicts
         if self.ice_controlling and 'ICE-CONTROLLING' in message.attributes:
             self.__log_info('Role conflict, expected to be controlling')
-            if self.tie_breaker >= message.attributes['ICE-CONTROLLING']:
+            if self._tie_breaker >= message.attributes['ICE-CONTROLLING']:
                 self.respond_error(message, addr, protocol, (487, 'Role Conflict'))
                 return
             self.switch_role(ice_controlling=False)
         elif not self.ice_controlling and 'ICE-CONTROLLED' in message.attributes:
             self.__log_info('Role conflict, expected to be controlled')
-            if self.tie_breaker < message.attributes['ICE-CONTROLLED']:
+            if self._tie_breaker < message.attributes['ICE-CONTROLLED']:
                 self.respond_error(message, addr, protocol, (487, 'Role Conflict'))
                 return
             self.switch_role(ice_controlling=True)
@@ -708,8 +811,8 @@ class Connection:
         response.add_fingerprint()
         protocol.send_stun(response, addr)
 
-        if not self.check_list:
-            self.early_checks.append((message, addr, protocol))
+        if not self._check_list:
+            self._early_checks.append((message, addr, protocol))
         else:
             self.check_incoming(message, addr, protocol)
 
@@ -724,15 +827,36 @@ class Connection:
         protocol.send_stun(response, addr)
 
     def sort_check_list(self):
-        sort_candidate_pairs(self.check_list, self.ice_controlling)
+        sort_candidate_pairs(self._check_list, self.ice_controlling)
 
     def switch_role(self, ice_controlling):
         self.__log_info('Switching to %s role', ice_controlling and 'controlling' or 'controlled')
         self.ice_controlling = ice_controlling
         self.sort_check_list()
 
+    def _unfreeze_initial(self):
+        # unfreeze first pair for the first component
+        first_pair = None
+        for pair in self._check_list:
+            if pair.component == min(self._components):
+                first_pair = pair
+                break
+        if first_pair is None:
+            return
+        if first_pair.state == CandidatePair.State.FROZEN:
+            self.check_state(first_pair, CandidatePair.State.WAITING)
+
+        # unfreeze pairs with same component but different foundations
+        seen_foundations = set(first_pair.local_candidate.foundation)
+        for pair in self._check_list:
+            if (pair.component == first_pair.component and
+               pair.local_candidate.foundation not in seen_foundations and
+               pair.state == CandidatePair.State.FROZEN):
+                self.check_state(pair, CandidatePair.State.WAITING)
+                seen_foundations.add(pair.local_candidate.foundation)
+
     def __log_info(self, msg, *args):
-        logger.info(repr(self) + ' ' + msg, *args)
+        logger.info('%s ' + msg, self, *args)
 
     def __repr__(self):
-        return 'Connection(%s)' % self.id
+        return 'Connection(%s)' % self._id
